//  (C) Copyright 2008 Anthony Williams 
//
//  Distributed under the Boost Software License, Version 1.0. (See
//  accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)

#include "boost/thread/thread.hpp"
#include "boost/thread/mutex.hpp"
#include "boost/thread/condition.hpp"
#include "n2561_future.hpp"
#include <iostream>
#include <assert.h>

int make_int()
{
    return 42;
}

int throw_runtime_error()
{
    throw std::runtime_error("42");
}

void set_promise_thread(jss::promise<int>* p)
{
    p->set_value(42);
}

struct my_exception
{};

void set_promise_exception_thread(jss::promise<int>* p)
{
    p->set_exception(boost::copy_exception(my_exception()));
}


void test_store_value_from_thread()
{
    jss::promise<int> pi2;
    jss::unique_future<int> fi2=pi2.get_future();
    boost::thread(set_promise_thread,&pi2);
    int j=fi2.get();
    assert(j==42);
    assert(fi2.is_ready());
    assert(fi2.has_value());
    assert(!fi2.has_exception());
    assert(!fi2.was_moved());
    assert(fi2.get_state()==jss::future_state::ready);
}


void test_store_exception()
{
    jss::promise<int> pi3;
    jss::unique_future<int> fi3=pi3.get_future();
    boost::thread(set_promise_exception_thread,&pi3);
    try
    {
        fi3.get();
        assert(false);
    }
    catch(my_exception)
    {
        assert(true);
    }
    
    assert(fi3.is_ready());
    assert(!fi3.has_value());
    assert(fi3.has_exception());
    assert(!fi3.was_moved());
    assert(fi3.get_state()==jss::future_state::ready);
}

void test_initial_state()
{
    jss::unique_future<int> fi;
    assert(!fi.is_ready());
    assert(!fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::uninitialized);
    int i;
    try
    {
        fi.try_move(i);
        assert(false);
    }
    catch(jss::future_uninitialized)
    {
        assert(true);
    }

    try
    {
        fi.try_get(i);
        assert(false);
    }
    catch(jss::future_uninitialized)
    {
        assert(true);
    }
}

void test_waiting_future()
{
    jss::promise<int> pi;
    jss::unique_future<int> fi;
    fi=pi.get_future();

    int i=0;
    assert(!fi.is_ready());
    assert(!fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::waiting);
    assert(!fi.try_move(i));
    assert(!fi.try_get(i));
    assert(i==0);
}

void test_cannot_get_future_twice()
{
    jss::promise<int> pi;
    pi.get_future();

    try
    {
        pi.get_future();
        assert(false);
    }
    catch(jss::future_already_retrieved&)
    {
        assert(true);
    }
}

void test_set_value_updates_future_state()
{
    jss::promise<int> pi;
    jss::unique_future<int> fi;
    fi=pi.get_future();

    pi.set_value(42);
    
    assert(fi.is_ready());
    assert(fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::ready);
}

void test_set_value_can_be_retrieved()
{
    jss::promise<int> pi;
    jss::unique_future<int> fi;
    fi=pi.get_future();

    pi.set_value(42);
    
    int i=0;
    assert(fi.try_get(i));
    assert(i==42);
    assert(fi.is_ready());
    assert(fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::ready);
}

void test_set_value_can_be_moved()
{
    jss::promise<int> pi;
    jss::unique_future<int> fi;
    fi=pi.get_future();

    pi.set_value(42);
    
    int i=0;
    assert(fi.try_move(i));
    assert(i==42);
    assert(!fi.is_ready());
    assert(!fi.has_value());
    assert(!fi.has_exception());
    assert(fi.was_moved());
    assert(fi.get_state()==jss::future_state::moved);
}

void test_future_from_packaged_task_is_waiting()
{
    jss::packaged_task<int> pt(make_int);
    jss::unique_future<int> fi=pt.get_future();
    int i=0;
    assert(!fi.is_ready());
    assert(!fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::waiting);
    assert(!fi.try_move(i));
    assert(!fi.try_get(i));
    assert(i==0);
}

void test_invoking_a_packaged_task_populates_future()
{
    jss::packaged_task<int> pt(make_int);
    jss::unique_future<int> fi=pt.get_future();

    pt();

    int i=0;
    assert(fi.is_ready());
    assert(fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::ready);
    assert(fi.try_get(i));
    assert(i==42);
}

void test_invoking_a_packaged_task_twice_throws()
{
    jss::packaged_task<int> pt(make_int);

    pt();
    try
    {
        pt();
        assert(false);
    }
    catch(jss::task_already_started)
    {
        assert(true);
    }
}


void test_cannot_get_future_twice_from_task()
{
    jss::packaged_task<int> pt(make_int);
    pt.get_future();
    try
    {
        pt.get_future();
        assert(false);
    }
    catch(jss::future_already_retrieved)
    {
        assert(true);
    }
}

void test_task_stores_exception_if_function_throws()
{
    jss::packaged_task<int> pt(throw_runtime_error);
    jss::unique_future<int> fi=pt.get_future();

    pt();

    int i=0;
    assert(fi.is_ready());
    assert(!fi.has_value());
    assert(fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::ready);
    try
    {
        fi.get();
        assert(false);
    }
    catch(std::exception&)
    {
        assert(true);
    }
    catch(...)
    {
        assert(!"Unknown exception thrown");
    }
    
}

void test_void_promise()
{
    jss::promise<void> p;
    jss::unique_future<void> f=p.get_future();
    p.set_value();
    assert(f.is_ready());
    assert(f.has_value());
    assert(!f.has_exception());
    assert(!f.was_moved());
    assert(f.get_state()==jss::future_state::ready);
    f.get();
    f.move();
    assert(f.get_state()==jss::future_state::moved);
}

void test_reference_promise()
{
    jss::promise<int&> p;
    jss::unique_future<int&> f=p.get_future();
    int i=42;
    p.set_value(i);
    assert(f.is_ready());
    assert(f.has_value());
    assert(!f.has_exception());
    assert(!f.was_moved());
    assert(f.get_state()==jss::future_state::ready);
    assert(&f.get()==&i);
    assert(&f.move()==&i);
    assert(f.get_state()==jss::future_state::moved);
}

void do_nothing()
{}

void test_task_returning_void()
{
    jss::packaged_task<void> pt(do_nothing);
    jss::unique_future<void> fi=pt.get_future();

    pt();

    assert(fi.is_ready());
    assert(fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::ready);
    fi.move();
    assert(fi.get_state()==jss::future_state::moved);
}

int global_ref_target=0;

int& return_ref()
{
    return global_ref_target;
}

void test_task_returning_reference()
{
    jss::packaged_task<int&> pt(return_ref);
    jss::unique_future<int&> fi=pt.get_future();

    pt();

    assert(fi.is_ready());
    assert(fi.has_value());
    assert(!fi.has_exception());
    assert(!fi.was_moved());
    assert(fi.get_state()==jss::future_state::ready);
    int& i=fi.move();
    assert(fi.get_state()==jss::future_state::moved);
    assert(&i==&global_ref_target);
}

void test_shared_future()
{
    jss::packaged_task<int> pt(make_int);
    jss::unique_future<int> fi=pt.get_future();

    jss::shared_future<int> sf(boost::move(fi));
    assert(fi.get_state()==jss::future_state::uninitialized);

    pt();

    int i=0;
    assert(sf.is_ready());
    assert(sf.has_value());
    assert(!sf.has_exception());
    assert(!sf.was_moved());
    assert(sf.get_state()==jss::future_state::ready);
    assert(sf.try_get(i));
    assert(i==42);
}

void test_copies_of_shared_future_become_ready_together()
{
    jss::packaged_task<int> pt(make_int);
    jss::unique_future<int> fi=pt.get_future();

    jss::shared_future<int> sf(boost::move(fi));
    jss::shared_future<int> sf2(sf);
    jss::shared_future<int> sf3;
    sf3=sf;
    assert(sf.get_state()==jss::future_state::waiting);
    assert(sf2.get_state()==jss::future_state::waiting);
    assert(sf3.get_state()==jss::future_state::waiting);

    pt();

    int i=0;
    assert(sf.is_ready());
    assert(sf.has_value());
    assert(!sf.has_exception());
    assert(!sf.was_moved());
    assert(sf.get_state()==jss::future_state::ready);
    assert(sf.try_get(i));
    assert(i==42);
    i=0;
    assert(sf2.is_ready());
    assert(sf2.has_value());
    assert(!sf2.has_exception());
    assert(!sf2.was_moved());
    assert(sf2.get_state()==jss::future_state::ready);
    assert(sf2.try_get(i));
    assert(i==42);
    i=0;
    assert(sf3.is_ready());
    assert(sf3.has_value());
    assert(!sf3.has_exception());
    assert(!sf3.was_moved());
    assert(sf3.get_state()==jss::future_state::ready);
    assert(sf3.try_get(i));
    assert(i==42);
}

void test_shared_future_can_be_move_assigned_from_unique_future()
{
    jss::packaged_task<int> pt(make_int);
    jss::unique_future<int> fi=pt.get_future();

    jss::shared_future<int> sf;
    sf=boost::move(fi);
    assert(fi.get_state()==jss::future_state::uninitialized);

    int i=0;
    assert(!sf.is_ready());
    assert(!sf.has_value());
    assert(!sf.has_exception());
    assert(!sf.was_moved());
    assert(sf.get_state()==jss::future_state::waiting);
    assert(!sf.try_get(i));
    assert(i==0);
}

void test_shared_future_void()
{
    jss::packaged_task<void> pt(do_nothing);
    jss::unique_future<void> fi=pt.get_future();

    jss::shared_future<void> sf(boost::move(fi));
    assert(fi.get_state()==jss::future_state::uninitialized);

    pt();

    assert(sf.is_ready());
    assert(sf.has_value());
    assert(!sf.has_exception());
    assert(!sf.was_moved());
    assert(sf.get_state()==jss::future_state::ready);
    sf.get();
}

void test_shared_future_ref()
{
    jss::promise<int&> p;
    jss::shared_future<int&> f(p.get_future());
    int i=42;
    p.set_value(i);
    assert(f.is_ready());
    assert(f.has_value());
    assert(!f.has_exception());
    assert(!f.was_moved());
    assert(f.get_state()==jss::future_state::ready);
    assert(&f.get()==&i);
}


int main()
{
    test_initial_state();
    test_waiting_future();
    test_cannot_get_future_twice();
    test_set_value_updates_future_state();
    test_set_value_can_be_retrieved();
    test_set_value_can_be_moved();
    test_store_value_from_thread();
    test_store_exception();
    test_future_from_packaged_task_is_waiting();
    test_invoking_a_packaged_task_populates_future();
    test_invoking_a_packaged_task_twice_throws();
    test_cannot_get_future_twice_from_task();
    test_task_stores_exception_if_function_throws();
    test_void_promise();
    test_reference_promise();
    test_task_returning_void();
    test_task_returning_reference();
    test_shared_future();
    test_copies_of_shared_future_become_ready_together();
    test_shared_future_can_be_move_assigned_from_unique_future();
    test_shared_future_void();
    test_shared_future_ref();
}
